__author__ = 'Calvin'
import weakref  # !

from pyperform import ComparisonBenchmark


class A(object):  # !
    def __init__(self):
        self.count = 0


a = A()  # !
ar = weakref.ref(a)  # !


@ComparisonBenchmark('Group1', largs=())
def without_weakref():
    a = A()
    for i in range(1000):
        a.count += 1
    return a.count


@ComparisonBenchmark('Group1', largs=())
def with_weakref():
    for i in range(1000):
        ar().count += 1
    return ar().count


ComparisonBenchmark.summarize('Group1')



__author__ = 'Calvin'
try:
    from builtins import range
except ImportError:
    range = xrange
from pyperform import *


class SomeClass(object): #!

    def __init__(self, n):
        self.n = n
        self.count = 0
        if n > 0:
            self.a = SomeClass(n-1)

    def func(self):
        self.count += 1

        return sum(range(10))

@BenchmarkedClass()
class MyClass(object):

    def __init__(self):
        self.obj = SomeClass(5)             # setup an object with some nested lookup
        self.g = self.generator()           # setup the generator in advance

    @ComparisonBenchmark('gen', classname='MyClass', validation=True)
    def call_generator(self):
        for i in self.g:                # Call the generator which calls the function 100 times (like not_generator)
            pass
        return self.obj.a.a.a.a.count

    def generator(self):
        func = self.obj.a.a.a.a.func
        for i in range(100):
            func()
            yield i


    @ComparisonBenchmark('gen', classname='MyClass', validation=True)
    def not_generator(self):
        func = self.obj.a.a.a.a.func
        for i in range(100):
            func()
        return self.obj.a.a.a.a.count


if __name__ == '__main__':

    # c = MyClass()
    # c.call_generator()

    with open('report.txt', 'w') as _f:
        ComparisonBenchmark.summarize('gen', _f, include_source=1)



__author__ = 'Calvin'
from pyperform import ComparisonBenchmark, BenchmarkedFunction


@ComparisonBenchmark('Group1', largs=(100,))
def mytest(l):
    out = 0.
    for i in range(l):
        out += i
    return out

@ComparisonBenchmark('Group1', largs=(100,))
def mytest2(l):
    out = 0.
    for i in range(l):
        out += i

    return out

@BenchmarkedFunction(largs=(5, 2, 10))
def TripleMultiply(a, b, c):
    result = a * b * c
    return result

with open('report.txt', 'w') as f:
    ComparisonBenchmark.summarize('Group1', f)




__author__ = 'clobo'

from collections import namedtuple  #!
from pyperform import ComparisonBenchmark

MyNamedTuple = namedtuple('ButtonStatus', ('name', 'age', 'gender'))  #!


class MyClass(object):  #!
    __slots__ = ('name', 'age', 'gender')

    def __init__(self, name, age, gender):
        self.name = name
        self.age = age
        self.gender = gender


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
                     Creation Speed Tests
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''


@ComparisonBenchmark('class')
def create_class():
    for i in xrange(1000):
        c = MyClass('calvin', 25, 'male')


@ComparisonBenchmark('class')
def create_namedtuple():
    for i in xrange(1000):
        c = MyNamedTuple('calvin', 25, 'male')


@ComparisonBenchmark('class')
def create_dict():
    for i in xrange(1000):
        c = {'name': 'calvin', 'age': 25, 'gender': 'male'}


'''''''''''''''''''''''''''''''''''''''''''''''''''''''''
                    Lookup Speed Tests
'''''''''''''''''''''''''''''''''''''''''''''''''''''''''


@ComparisonBenchmark('lookup')
def class_attr_lookup():
    c = MyClass('calvin', 25, 'male')
    for i in xrange(10000):
        name, age, gender = c.name, c.age, c.gender


@ComparisonBenchmark('lookup')
def namedtuple_attr_lookup():
    c = MyNamedTuple('calvin', 25, 'male')
    for i in xrange(10000):
        name, age, gender = c.name, c.age, c.gender


@ComparisonBenchmark('lookup')
def dict_attr_lookup():
    c = {'name': 'calvin', 'age': 25, 'gender': 'male'}
    for i in xrange(10000):
        name, age, gender = c['name'], c['age'], c['gender']


ComparisonBenchmark.summarize('class')
ComparisonBenchmark.summarize('lookup')



__author__ = 'calvin'

import logging
from pyperform import StringIO
from .benchmark import Benchmark
from .tools import *
from .exceptions import ValidationError


class ComparisonBenchmark(Benchmark):
    groups = {}

    def __init__(self, group, classname=None, setup=None, validation=False, validation_func=None, largs=None, kwargs=None, **kw):
        super(ComparisonBenchmark, self).__init__(setup=setup, largs=largs, kwargs=kwargs, **kw)
        self.group = group
        self.classname = classname
        self.result_validation = validation
        self.validation_func = validation_func
        self.result = None
        if group not in self.groups:
            self.groups[group] = []

    def __call__(self, caller):
        if self.enable:
            super(ComparisonBenchmark, self).__call__(caller)
            self.groups[self.group].append(self)
            # Bound functions are tested in ClassBenchmark.__call__
            # Just store a reference to the ComparisonBenchmark if the function is bound, otherwise, run the test
            if not self.is_class_method:
                # Run the test
                self.run_timeit(self.stmt, self.setup_src)
                if self.result_validation:
                    self.validate()
        return caller

    def validate(self):
        """
        Execute the code once to get it's results (to be used in function validation). Compare the result to the
        first function in the group.
        """
        validation_code = self.setup_src + '\nvalidation_result = ' + self.stmt
        validation_scope = {}
        exec(validation_code, validation_scope)
        # Store the result in the first function in the group.
        if len(self.groups[self.group]) == 1:
            self.result = validation_scope['validation_result']
            logging.info('PyPerform: Validating group "{b.group}" against function "{b.callable.__name__}"'
                         .format(b=self))
        else:
            compare_against_benchmark = self.groups[self.group][0]
            test = [benchmark.result_validation for benchmark in self.groups[self.group]]
            if not all(test):
                raise ValueError('All functions within a group must have the same validation flag.')
            compare_result = compare_against_benchmark.result
            if self.validation_func:
                results_are_valid = self.validation_func(compare_result, validation_scope['validation_result'])
            else:
                results_are_valid = compare_result == validation_scope['validation_result']
            if results_are_valid:
                logging.info('PyPerform: Validating {}......PASSED!'.format(self.callable.__name__))
            else:
                error = 'Results of functions {0} and {1} are not equivalent.\n{0}:\t {2}\n{1}:\t{3}'
                raise ValidationError(error.format(compare_against_benchmark.callable.__name__, self.callable.__name__,
                                          compare_result, validation_scope['validation_result']))

    @staticmethod
    def summarize(group, fs=None, include_source=True):
        """
        Tabulate and write the results of ComparisonBenchmarks to a file or standard out.
        :param str group: name of the comparison group.
        :param fs: file-like object (Optional)
        """
        _line_break = '{0:-<120}\n'.format('')
        tests = sorted(ComparisonBenchmark.groups[group], key=lambda t: getattr(t, 'time_average_seconds'))
        log = StringIO.StringIO()
        log.write('Call statement:\n\n')
        log.write('\t' + tests[0].stmt)
        log.write('\n\n\n')
        fmt = "{0: <8} {1: <35} {2: <12} {3: <15} {4: <15} {5: <14}\n"
        log.write(fmt.format('Rank', 'Function Name', 'Time', '% of Slowest', 'timeit_repeat', 'timeit_number'))
        log.write(_line_break)
        log.write('\n')

        for i, t in enumerate(tests):
            func_name = "{}.{}".format(t.classname, t.callable.__name__) if t.classname else t.callable.__name__
            if i == len(tests)-1:
                time_percent = 'Slowest'
            else:
                time_percent = "{:.1f}".format(t.time_average_seconds / tests[-1].time_average_seconds * 100)
            log.write(fmt.format(i+1,
                                 func_name,
                                 convert_time_units(t.time_average_seconds),
                                 time_percent,
                                 t.timeit_repeat,
                                 t.timeit_number))
        log.write(_line_break)

        if include_source:
            log.write('\n\n\nSource Code:\n')
            log.write(_line_break)
            for test in tests:
                log.write(test.log.getvalue())
                log.write(_line_break)

        if isinstance(fs, str):
            with open(fs, 'w') as f:
                f.write(log.getvalue())

        elif fs is None:
            print(log.getvalue())
        else:
            try:
                fs.write(log.getvalue())
            except AttributeError as e:
                print(e)



__author__ = 'calvin'

import cProfile
import logging
from pyperform import StringIO
import os
import pstats
import sys
import threading

Thread = threading.Thread          # Start off using threading.Thread until changed
BaseThread = threading.Thread      # Store the Thread class from the threading module before monkey-patching
profiled_thread_enabled = False
logged_thread_enabled = True


def enable_thread_profiling(profile_dir, exception_callback=None):
    """
    Monkey-patch the threading.Thread class with our own ProfiledThread. Any subsequent imports of threading.Thread
    will reference ProfiledThread instead.
    """
    global profiled_thread_enabled, Thread
    if os.path.isdir(profile_dir):
        ProfiledThread.profile_dir = profile_dir
    else:
        raise OSError('%s does not exist' % profile_dir)
    ProfiledThread.exception_callback = exception_callback
    Thread = threading.Thread = ProfiledThread
    profiled_thread_enabled = True


def enable_thread_logging(exception_callback=None):
    """
    Monkey-patch the threading.Thread class with our own LoggedThread. Any subsequent imports of threading.Thread
    will reference LoggedThread instead.
    """
    global logged_thread_enabled, Thread
    LoggedThread.exception_callback = exception_callback
    Thread = threading.Thread = LoggedThread
    logged_thread_enabled = True


class ProfiledThread(BaseThread):
    """
    A Thread that contains it's own profiler. When the SSI_App closes, all profiles are combined and printed
    to a single .profile.
    """
    profile_dir = None
    exception_callback = None

    def run(self):
        profiler = cProfile.Profile()
        try:
            logging.debug('ProfiledThread: Starting ProfiledThread {}'.format(self.name))
            profiler.runcall(BaseThread.run, self)
        except Exception as e:
            logging.error('ProfiledThread: Error encountered in Thread {name}'.format(name=self.name))
            logging.error(e)
            if ProfiledThread.exception_callback:
                e_type, e_value, last_traceback = sys.exc_info()
                ProfiledThread.exception_callback(e_type, e_value, last_traceback)
        finally:
            if ProfiledThread.profile_dir is None:
                logging.warning('ProfiledThread: profile_dir is not specified. '
                                'Profile \'{}\' will not be saved.'.format(self.name))
                return
            self.print_stats(profiler)

    def print_stats(self, profiler):
        filename = os.path.join(self.profile_dir, self.name)
        s = StringIO.StringIO()
        sortby = 'cumulative'
        ps = pstats.Stats(profiler, stream=s)
        # Take out directory names
        ps.strip_dirs()
        # Sort
        ps.sort_stats(sortby)
        # Print to the stream
        ps.print_stats()

        stats_file = filename + '.stats'
        profile_file = filename + '.profile'
        # Create the stats file
        ps.dump_stats(stats_file)
        # Create a readable .profile file
        with open(profile_file, 'w') as f:
            f.write(s.getvalue())

    @staticmethod
    def combine_profiles(profile_dir, outfile, sortby='cumulative'):
        s = StringIO.StringIO()
        stat_files = [f for f in os.listdir(profile_dir) if os.path.isfile(os.path.join(profile_dir, f))
                      and f.endswith('.stats')]
        ps = pstats.Stats(os.path.join(profile_dir, stat_files[0]), stream=s)
        if len(stat_files) > 1:
            for stat in stat_files[1:]:
                ps.add(os.path.join(profile_dir, stat))

        profile_name = os.path.join(profile_dir, '{}.profile'.format(outfile.replace('.profile', '')))
        with open(profile_name, 'w') as f:
            ps.strip_dirs()
            ps.sort_stats(sortby)
            ps.print_stats()
            f.write(s.getvalue())


class LoggedThread(BaseThread):

    exception_callback = None

    def run(self):
        logging.debug('LoggedThread: Starting LoggedThread {}'.format(self.name))
        try:
            super(LoggedThread, self).run()
        except Exception as e:
            logging.error('LoggedThread: Error encountered in Thread {name}'.format(name=self.name))
            logging.error(e)
            if LoggedThread.exception_callback:
                e_type, e_value, last_traceback = sys.exc_info()
                LoggedThread.exception_callback(e_type, e_value, last_traceback)





from __future__ import print_function

__version__ = '1.86'
import sys

if sys.version[0] == '3':
    import io as StringIO  # Python 3.x
else:
    import cStringIO as StringIO  # Python 2.x

    range = xrange

from pyperform.benchmark import Benchmark
from .comparisonbenchmark import ComparisonBenchmark
from .benchmarkedclass import BenchmarkedClass
from .benchmarkedfunction import BenchmarkedFunction
from .thread import Thread
from .timer import timer
from .exceptions import ValidationError
from .customlogger import CustomLogLevel, new_log_level


def enable():
    """
    Enable all benchmarking.
    """
    Benchmark.enable = True
    ComparisonBenchmark.enable = True
    BenchmarkedFunction.enable = True
    BenchmarkedClass.enable = True


def disable():
    """
    Disable all benchmarking.
    """
    Benchmark.enable = False
    ComparisonBenchmark.enable = False
    BenchmarkedFunction.enable = False
    BenchmarkedClass.enable = False














__author__ = 'calvin'

from time import time

from .tools import convert_time_units

timer_format = "{name}: recent: {recent_time} average: {avg_time}"


def timer(func):
    def _timed_function(*args, **kwargs):
        t1 = time()
        result = func(*args, **kwargs)
        t2 = time()
        delta = t2-t1
        _timed_function._total_time += delta
        _timed_function._call_count += 1
        _timed_function._average_time = convert_time_units(_timed_function._total_time / _timed_function._call_count)
        print(timer_format.format(name=func.__name__, recent_time=convert_time_units(delta), avg_time=_timed_function._average_time))
        return result

    _timed_function._total_time = 0
    _timed_function._call_count = 0
    _timed_function.__name__ = "_timed_{}".format(func.__name__)

    return _timed_function



__author__ = 'clobo'

import logging


class CustomLogLevel(object):

    def __init__(self, level, name, logger_name=None):
        self.level = level
        self.name = name
        self.logger_name = logger_name
        if logger_name is None:
            self.logger = logging.getLogger()
        else:
            self.logger = logging.getLogger(logger_name)

    def __call__(self, customlevel):
        """
        Wrap the decorated function to take care of the setting up of the custom log level.

        """
        # Add the new custom level to the list of known levels
        logging.addLevelName(self.level, self.name)

        def _wrapper(msg, *args, **kwargs):
            # Check if the currently set level allows this log level to print.
            if self.logger.isEnabledFor(level):
                _msg, _args, _kwargs = customlevel(self.logger, msg, *args, **kwargs)
                self.logger.log(level, _msg, *_args, **_kwargs)

        # Create function bindings in the logger or if using the root logger, setup the bindings to allow
        # calls to logging.mycustomlevel() much like logging.info(), logging.debug() etc.
        setattr(self.logger, self.name.lower(), _wrapper)
        if self.logger_name is None:
            setattr(logging, self.name.lower(), _wrapper)

        return customlevel


def new_log_level(level, name, logger_name=None):
    """
    Quick way to create a custom log level that behaves like the default levels in the logging module.
    :param level: level number
    :param name: level name
    :param logger_name: optional logger name
    """
    @CustomLogLevel(level, name, logger_name)
    def _default_template(logger, msg, *args, **kwargs):
        return msg, args, kwargs


if __name__ == '__main__':

    level = logging.INFO-5
    name = 'MYLEVEL'
    # logger_name = 'mylogger'
    logger_name = None

    @CustomLogLevel(level, name, logger_name=logger_name)
    def myloglevel(logger, msg, *args, **kwargs):
        return 'This is a custom level: %s' % msg, args, kwargs

    # create_new_level(level, name, logger_name=logger_name)

    logging.basicConfig()
    if logger_name:
        l = logging.getLogger(logger_name)
        logger = l
    else:
        l = logging.getLogger()
        logger = logging

    l.setLevel(logging.INFO)
    logger.info('this is a test')
    logger.mylevel('this is a test')
    l.setLevel(level)
    logger.info('this is a test')
    logger.mylevel('this is a test')





__author__ = 'Calvin'
"""
This example demonstrates how pyperform can be used to benchmark class functions. In this example we use
ComparisonBenchmarks to compare the speed of two methods which calculates a person's savings.

** Note that when benchmarking class methods, the classname argument to ComparisonBenchmark must be provided.

The class, Person, is initialized with several required parameters (name, age, monthly_income) and some optional
parameters (height). The two methods to calculate savings both have a required argument (retirement_age) and an optional
argument of monthly spending (monthly_spending).

"""

from pyperform import BenchmarkedClass, ComparisonBenchmark, BenchmarkedFunction

@BenchmarkedClass(largs=('Calvin', 24, 1000.,), kwargs={'height': '165 cm'})
class Person(object):

    def __init__(self, name, age, monthly_income, height=None, *args, **kwargs):
        self.name = name
        self.age = age
        self.height = height
        self.monthly_income = monthly_income

    @ComparisonBenchmark('Calculate Savings', classname="Person", timeit_number=100, validation=True, largs=(55,),
                         kwargs={'monthly_spending': 500})
    def calculate_savings_method1(self, retirement_age, monthly_spending=0):
        savings = 0
        for y in range(self.age, retirement_age):
            for m in range(12):
                savings += self.monthly_income - monthly_spending
        return savings

    @ComparisonBenchmark('Calculate Savings', classname="Person", timeit_number=100, validation=True, largs=(55,),
                         kwargs={'monthly_spending': 500})
    def calculate_savings_method2(self, retirement_age, monthly_spending=0):
        yearly_income = 12 * (self.monthly_income - monthly_spending)
        n_years = retirement_age - self.age
        if n_years > 0:
            return yearly_income * n_years

    @BenchmarkedFunction(classname="Person", timeit_number=100, largs=(55,), kwargs={'monthly_spending': 500})
    def same_as_method_2(self, retirement_age, monthly_spending=0):
        yearly_income = 12 * (self.monthly_income - monthly_spending)
        n_years = retirement_age - self.age
        if n_years > 0:
            return yearly_income * n_years


# Can print the summary to file or if ComparisonBenchmark.summarize() is not given an fs parameter, it will print to
# console.
report_file = open('report.txt', 'w')
ComparisonBenchmark.summarize(group='Calculate Savings', fs=report_file)


__author__ = 'calvin'


from pyperform import *


@ComparisonBenchmark('imports', timeit_repeat=10)
def import_top_level():
    import os
    import datetime
    import collections
    import json


@ComparisonBenchmark('imports', timeit_repeat=10)
def import_lower_level():
    from os import path
    from datetime import datetime
    from collections import deque
    from json import decoder




ComparisonBenchmark.summarize('imports')



__author__ = 'calvin'

import logging

from .benchmark import Benchmark
from .benchmarkedfunction import BenchmarkedFunction
from .tools import convert_time_units
from .exceptions import ValidationError


class BenchmarkedClass(Benchmark):
    bound_functions = {}

    def __init__(self, setup=None, largs=None, kwargs=None, **kw):
        super(BenchmarkedClass, self).__init__(setup, largs=largs, kwargs=kwargs, **kw)

    def __call__(self, cls):
        if self.enable:
            super(BenchmarkedClass, self).__call__(cls)
            setup_src = self.setup_src
            setup_src += '\ninstance = {}'.format(self.stmt)

            groups = set()
            for p in self.bound_functions[cls.__name__]:
                stmt = p.stmt
                p.run_timeit(stmt, setup_src)
                p.write_log()
                if isinstance(p, BenchmarkedFunction):
                    print("{} \t {}".format(p.callable.__name__, convert_time_units(p.time_average_seconds)))
                if hasattr(p, 'result_validation') and p.result_validation and p.group not in groups:
                    self.validate(p.groups[p.group])
                groups.add(p.group)

        return cls

    def validate(self, benchmarks):
        """
        Execute the code once to get it's results (to be used in function validation). Compare the result to the
        first function in the group.
        :param benchmarks: list of benchmarks to validate.
        """
        class_code = self.setup_src
        instance_creation = '\ninstance = {}'.format(self.stmt)
        for i, benchmark in enumerate(benchmarks):
            if not benchmark.result_validation:
                break

            validation_code = class_code + instance_creation + '\nvalidation_result = ' + benchmark.stmt
            validation_scope = {}
            exec(validation_code, validation_scope)
            # Store the result in the first function in the group.
            if i == 0:
                compare_against_function = benchmarks[0].callable.__name__
                compare_against_result = validation_scope['validation_result']
                logging.info('PyPerform: Validating group "{b.group}" against method '
                             '"{b.classname}.{b.callable.__name__}"'.format(b=benchmarks[0]))
            else:
                if compare_against_result == validation_scope['validation_result']:
                    logging.info('PyPerform: Validating {b.classname}.{b.callable.__name__}......PASSED!'
                                 .format(b=benchmark))
                else:
                    error = 'Results of functions {0} and {1} are not equivalent.\n{0}:\t {2}\n{1}:\t{3}'
                    raise ValidationError(error.format(compare_against_function, benchmark.callable.__name__,
                                              compare_against_result, validation_scope['validation_result']))



__author__ = 'calvin'
"""
This example demonstrates the performance benefit of using list comprehension instead of for loop and append.
It also demonstrates how to use imported modules in your benchmarks as well as compare functions of the same group.
"""

try:
    from builtins import range
except ImportError:
    range = xrange
from math import sin  #!

from pyperform import ComparisonBenchmark


@ComparisonBenchmark('Group1', validation=True, largs=(100,))
def list_append(n, *args, **kwargs):
    l = []
    for i in range(1, n):
        l.append(sin(i))
    return l


@ComparisonBenchmark('Group1', validation=True, largs=(100,))
def list_comprehension(n, *args, **kwargs):
    l = [sin(i) for i in range(1, n)]
    return l


ComparisonBenchmark.summarize('Group1', fs='report.txt')



__author__ = 'calvin'

from .benchmark import Benchmark
from .tools import convert_time_units

class BenchmarkedFunction(Benchmark):

    def __call__(self, caller):
        if self.enable:
            super(BenchmarkedFunction, self).__call__(caller)
            if not self.is_class_method:
                self.run_timeit(self.stmt, self.setup_src)
                print("{} \t {}".format(caller.__name__, convert_time_units(self.time_average_seconds)))
        return caller



__author__ = 'calvin'

import sys
import re
from math import log10

if sys.version[0] == '3':
    pass
else:
    range = xrange


classdef_regex = re.compile(r"\S*def .*#!|class .*#!")
tagged_line_regex = re.compile(r".*#!")


def convert_time_units(t):
    """ Convert time in seconds into reasonable time units. """
    order = log10(t)
    if -9 < order < -6:
        time_units = 'ns'
        factor = 1000000000
    elif -6 < order < -3:
        time_units = 'us'
        factor = 1000000
    elif -3 <= order < -1:
        time_units = 'ms'
        factor = 1000.
    elif -1 <= order:
        time_units = 's'
        factor = 1
    return "{:.3f} {}".format(factor * t, time_units)


def globalize_indentation(src):
    """ Strip the indentation level so the code runs in the global scope. """
    lines = src.splitlines()
    indent = len(lines[0]) - len(lines[0].strip(' '))
    func_src = ''
    for ii, l in enumerate(src.splitlines()):
        line = l[indent:]
        func_src += line + '\n'
    return func_src


def remove_decorators(src):
    """ Remove decorators from the source code """
    src = src.strip()
    src_lines = src.splitlines()
    multi_line = False
    n_deleted = 0
    for n in range(len(src_lines)):
        line = src_lines[n - n_deleted].strip()
        if (line.startswith('@') and 'Benchmark' in line) or multi_line:
            del src_lines[n - n_deleted]
            n_deleted += 1
            if line.endswith(')'):
                multi_line = False
            else:
                multi_line = True
    setup_src = '\n'.join(src_lines)
    return setup_src


def get_tagged_imports(fp):
    imports = []
    inside_def = False
    def_lines = []
    def_indent = 0
    with open(fp, 'r') as f:
        lastLine = f.readline()
        for line in f:

            tagged_class_or_def = re.findall(classdef_regex, lastLine)
            tagged_line = re.findall(tagged_line_regex, lastLine)

            # Find the indentation level of the function/class definition and capture all source code lines
            # until we get a line that is the same indentation level (end of function/class definition).
            if tagged_class_or_def or inside_def:
                if tagged_class_or_def and def_lines:
                    imports.append(''.join(def_lines))
                    def_lines = []
                    inside_def = False

                if inside_def:
                    # For lines within the definition
                    indent = len(lastLine) - len(lastLine.lstrip(' '))
                    if indent == def_indent and lastLine != '\n':
                        imports.append(''.join(def_lines))
                        def_lines = []
                        inside_def = False
                        def_indent = 0
                        if tagged_line:
                            imports.append(lastLine)
                    else:
                        if lastLine != '\n':
                            def_lines.append(lastLine)
                else:
                    # For the definition line
                    inside_def = True
                    def_indent = len(lastLine) - len(lastLine.lstrip(' '))
                    def_lines.append(lastLine)
            elif tagged_line:
                imports.append(lastLine)

            lastLine = line

    # Examine the last line
    tagged_line = re.findall(tagged_line_regex, lastLine)
    if inside_def:
        def_lines.append(line)
        imports.append(''.join(def_lines))
    elif tagged_line:
        imports.append(line)

    src = '\n'.join(imports) + '\n'
    return src


def generate_call_statement(func, is_class_method, *args, **kwargs):
    # Create the call statement
    if is_class_method:
        stmt = 'instance.' + func.__name__ + '('
    else:
        stmt = func.__name__ + '('
    for arg in args:
        stmt += arg.__repr__() + ', '
    for kw, val in kwargs.items():
        stmt += '{0}={1}, '.format(kw, val.__repr__())
    stmt = stmt.strip(', ')
    stmt += ')'
    return stmt



__author__ = 'calvin'

import inspect
import timeit
from types import FunctionType
from pyperform import StringIO
from .tools import *


class Benchmark(object):
    enable = True

    def __init__(self, setup=None, classname=None, timeit_repeat=3, timeit_number=1000, largs=None, kwargs=None):
        self.setup = setup
        self.timeit_repeat = timeit_repeat
        self.timeit_number = timeit_number
        self.classname = classname
        self.group = None
        self.is_class_method = None
        if largs is not None and type(largs) is tuple:
            self._args = largs[:]
        else:
            self._args = ()
        self._kwargs = kwargs.copy() if kwargs is not None else {}
        self.setup_src = ''
        self.callable = None
        self._is_function = None
        self.log = StringIO.StringIO()
        self.time_average_seconds = None

    def __call__(self, caller):
        if self.enable:
            self.callable = caller
            self._is_function = isinstance(caller, FunctionType)

            fp = inspect.getfile(caller)
            imports = get_tagged_imports(fp)
            func_src = remove_decorators(globalize_indentation(inspect.getsource(caller)))

            # Determine if the function is bound. If it is, keep track of it so we can run the benchmark after the class
            # benchmark has been initialized.
            src_lines = func_src.splitlines()
            self.is_class_method = 'def' in src_lines[0] and 'self' in src_lines[0]
            if self.is_class_method and self.classname:
                from .benchmarkedclass import BenchmarkedClass

                try:
                    BenchmarkedClass.bound_functions[self.classname].append(self)
                except KeyError:
                    BenchmarkedClass.bound_functions[self.classname] = [self]

            if callable(self.setup):
                setup_func = inspect.getsource(self.setup)
                setup_src = globalize_indentation(setup_func[setup_func.index('\n') + 1:])
            elif type(self.setup) == str:
                setup_src = self.setup
            else:
                setup_src = ''

            src = '\n'.join([imports, setup_src, func_src])
            self.setup_src = src + '\n'
            self.log.write(self.setup_src)

            self.stmt = generate_call_statement(caller, self.is_class_method, *self._args, **self._kwargs)

        return caller

    def write_log(self, fs=None):
        """
        Write the results of the benchmark to a log file.
        :param fs: file-like object.
        """
        log = StringIO.StringIO()
        log.write(self.setup_src)

        # If the function is not bound, write the test score to the log
        if not self.is_class_method:
            time_avg = convert_time_units(self.time_average_seconds)
            log.write("\nAverage time: {0} \n".format(time_avg))

        if fs:
            with open(fs, 'w') as _f:
                _f.write(log.getvalue())

    def run_timeit(self, stmt, setup):
        """ Create the function call statement as a string used for timeit. """
        _timer = timeit.Timer(stmt=stmt, setup=setup)
        trials = _timer.repeat(self.timeit_repeat, self.timeit_number)
        self.time_average_seconds = sum(trials) / len(trials) / self.timeit_number
        # Convert into reasonable time units
        time_avg = convert_time_units(self.time_average_seconds)
        return time_avg



__author__ = 'calvin'
import random #!

from pyperform import ComparisonBenchmark


def _setup():
    l = []
    lookup = {ii: l.append for ii in range(5)}


largs = (random.randint(0, 4),)

@ComparisonBenchmark('group1', setup=_setup, validation=True, timeit_number=10, timeit_repeat=1000)
def do_if():
    ii = random.randint(0, 4)
    if ii == 0:
        l.append(ii)
    elif ii == 1:
        l.append(ii)
    elif ii == 2:
        l.append(ii)
    elif ii == 3:
        l.append(ii)
    elif ii == 4:
        l.append(ii)

@ComparisonBenchmark('group1', setup=_setup, validation=True, timeit_number=10, timeit_repeat=1000)
def do_dict_lookup():
    ii = random.randint(0, 4)
    lookup[ii](ii)


ComparisonBenchmark.summarize('group1')


__author__ = 'lobocv'

from distutils.core import setup

from pyperform import __version__

setup(
    name='pyperform',
    packages=['pyperform'],  # this must be the same as the name above
    version=__version__,
    description='A fast and convenient way to performance test functions and compare results.',
    author='Calvin Lobo',
    author_email='calvinvlobo@gmail.com',
    url='https://github.com/lobocv/pyperform',
    download_url='https://github.com/lobocv/pyperform/tarball/%s' % __version__,
    keywords=['testing', 'performance', 'comparison', 'convenience', 'logging', 'timeit', 'speed', 'crash reporting'],
    classifiers=[],
)
